package com.flyfish.framework.config;

import com.flyfish.framework.config.properties.SecurityProperties;
import com.flyfish.framework.configuration.jwt.JwtProperties;
import com.flyfish.framework.configuration.jwt.JwtSecurityContextRepository;
import com.flyfish.framework.configuration.jwt.TokenProvider;
import com.flyfish.framework.domain.UserQo;
import com.flyfish.framework.domain.po.User;
import com.flyfish.framework.enums.UserStatus;
import com.flyfish.framework.enums.UserType;
import com.flyfish.framework.handler.JsonAuthenticationFailureHandler;
import com.flyfish.framework.handler.JsonAuthenticationSuccessHandler;
import com.flyfish.framework.handler.JsonLogoutSuccessHandler;
import com.flyfish.framework.initializer.UserInitializer;
import com.flyfish.framework.service.AuthenticationAuditor;
import com.flyfish.framework.service.AuthenticationLogger;
import com.flyfish.framework.service.UserService;
import org.springframework.boot.autoconfigure.condition.ConditionalOnMissingBean;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.boot.context.properties.EnableConfigurationProperties;
import org.springframework.context.annotation.Bean;
import org.springframework.core.annotation.Order;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.security.authentication.ReactiveAuthenticationManager;
import org.springframework.security.config.annotation.web.reactive.EnableWebFluxSecurity;
import org.springframework.security.config.web.server.ServerHttpSecurity;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.userdetails.ReactiveUserDetailsService;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.security.crypto.factory.PasswordEncoderFactories;
import org.springframework.security.crypto.password.DelegatingPasswordEncoder;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.security.web.server.SecurityWebFilterChain;
import org.springframework.security.web.server.authentication.HttpStatusServerEntryPoint;
import org.springframework.security.web.server.context.ServerSecurityContextRepository;
import org.springframework.security.web.server.context.WebSessionServerSecurityContextRepository;
import org.springframework.util.MultiValueMap;

import java.util.stream.Stream;

import static org.springframework.security.web.server.util.matcher.ServerWebExchangeMatchers.pathMatchers;

/**
 * @author wangyu
 */
@EnableWebFluxSecurity
@Order(1)
@EnableConfigurationProperties({JwtProperties.class, SecurityProperties.class})
public class WebSecurityConfig {

    /**
     * 设置密码加密策略
     *
     * @return 密码加密器
     */
    @Bean
    public static PasswordEncoder passwordEncoder() {
        DelegatingPasswordEncoder delegatingPasswordEncoder =
                (DelegatingPasswordEncoder) PasswordEncoderFactories.createDelegatingPasswordEncoder();
        //设置defaultPasswordEncoderForMatches为NoOpPasswordEncoder
        delegatingPasswordEncoder.setDefaultPasswordEncoderForMatches(new BCryptPasswordEncoder());
        return delegatingPasswordEncoder;
    }

    @Bean
    public AuthenticationAuditor loginAuditor() {
        return new AuthenticationAuditorImpl();
    }

    @Bean
    @ConditionalOnMissingBean(AuthenticationLogger.class)
    public AuthenticationLogger authenticationLogger() {
        return new AuthenticationLogger() {
            @Override
            public void success(UserDetails user) {

            }

            @Override
            public void failure(MultiValueMap<String, String> form, AuthenticationException exception) {

            }

            @Override
            public void logout(UserDetails user) {

            }
        };
    }

    @ConditionalOnProperty(value = "jwt.enable", havingValue = "true")
    @Bean("contextRepository")
    public JwtSecurityContextRepository jwtSecurityContextRepository() {
        return new JwtSecurityContextRepository();
    }

    @ConditionalOnProperty(value = "jwt.enable", havingValue = "false")
    @Bean
    public ServerSecurityContextRepository contextRepository() {
        return new WebSessionServerSecurityContextRepository();
    }

    @Bean
    public TokenProvider tokenProvider(JwtProperties properties) {
        return new TokenProvider(properties);
    }

    /**
     * spring安全拦截规则配置
     *
     * @param http            配置器
     * @param dataTransformer 数据转换器
     * @return 结果
     */
    @Bean
    @SuppressWarnings("all")
    public SecurityWebFilterChain springSecurityFilterChain(ServerHttpSecurity http,
                                                            TokenProvider tokenProvider,
                                                            SecurityProperties properties,
                                                            ReactiveUserDetailsService userDetailsService,
                                                            AuthenticationAuditor authenticationAuditor) {
        http
                .securityContextRepository(contextRepository())
                .authorizeExchange()
                .pathMatchers(Stream.concat(Stream.of(properties.getAllowUris()), Stream.of("/api/logout", "/api/login"))
                        .toArray(String[]::new)).permitAll()
                .pathMatchers("/api/users/**").authenticated()
                .anyExchange().authenticated()
                .and()
                .formLogin() // 配置登录节点
                .authenticationManager(authenticationManager(properties, userDetailsService))
                .authenticationEntryPoint(new HttpStatusServerEntryPoint(HttpStatus.UNAUTHORIZED))
                .authenticationSuccessHandler(new JsonAuthenticationSuccessHandler(authenticationAuditor))
                .authenticationFailureHandler(new JsonAuthenticationFailureHandler(authenticationAuditor))
                .requiresAuthenticationMatcher(pathMatchers(HttpMethod.POST, "/login", "/api/login"))
                .and()
                .logout()
                .logoutUrl("/api/logout")
                .logoutSuccessHandler(new JsonLogoutSuccessHandler(authenticationAuditor, tokenProvider))
                .and()
                .csrf().disable();
        return http.build();
    }

    /**
     * 构建鉴权管理器
     *
     * @param userDetailsService 用户详情服务
     * @return 结果
     */
    public ReactiveAuthenticationManager authenticationManager(SecurityProperties securityProperties,
                                                               ReactiveUserDetailsService userDetailsService) {
        RSAAuthenticationManager authenticationManager = new RSAAuthenticationManager(securityProperties, userDetailsService);
        authenticationManager.setPasswordEncoder(passwordEncoder());
        return authenticationManager;
    }

    /**
     * 默认的初始化
     *
     * @param userService 用户服务
     * @return 结果
     */
    @Bean
    @ConditionalOnMissingBean(UserInitializer.class)
    public UserInitializer userInitializer(UserService userService) {
        return () -> {
            UserQo qo = new UserQo();
            qo.setType(UserType.SUPER_ADMIN.name());
            if (userService.count(qo) == 0) {
                // 初始化用户
                User user = new User();
                user.setUsername("admin");
                user.setPassword("Unicom@2020");
                user.setType(UserType.SUPER_ADMIN);
                user.setEnable(true);
                user.setApp(false);
                user.setStatus(UserStatus.NORMAL);
                user.setPhone("10000000000");
                user.setName("超级管理员");
                user.setCode("Administrator");
                userService.createSelective(user);
            }
        };
    }
}
