package com.flyfish.framework.beans.meta;

import com.flyfish.framework.annotations.DictValue;
import com.flyfish.framework.annotations.Property;
import com.flyfish.framework.utils.StringFormats;
import lombok.Data;
import lombok.val;
import org.apache.commons.lang3.ClassUtils;
import org.apache.commons.lang3.reflect.FieldUtils;
import org.apache.commons.lang3.reflect.TypeUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.data.mongodb.core.mapping.Document;

import java.beans.PropertyDescriptor;
import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.*;
import java.util.stream.Collectors;

/**
 * bean的属性描述
 * 对应js的类型
 *
 * @author wangyu
 */
@Data
public class BeanProperty {

    // 属性名称
    private String name;

    // 属性标题
    private String title;

    // 描述信息
    private String description;

    // bean属性的类型，js类型
    private BeanPropertyType type;

    // 只读属性，不生成于表单
    private boolean readonly;

    // 属性
    private Map<String, Object> props;

    // 类型为object时，拥有子表单
    private List<BeanProperty> children;

    /**
     * 来自属性解释器构造
     * 支持对象嵌套
     *
     * @param descriptor 解释器
     * @return 结果
     */
    public static BeanProperty form(PropertyDescriptor descriptor, Class<?> beanClass) {
        BeanProperty property = new BeanProperty();
        property.setName(descriptor.getName());
        property.setType(BeanPropertyType.of(descriptor, beanClass));
        // 尝试获取field
        Field field = FieldUtils.getField(beanClass, descriptor.getName(), true);
        // 存在field，可以干一些坏事
        if (null != field) {
            Property props = AnnotationUtils.findAnnotation(field, Property.class);
            if (null != props) {
                String parentName = Optional.ofNullable(beanClass.getAnnotation(RestBean.class)).map(RestBean::name).orElse("");
                property.setTitle(props.inherited() ? parentName + props.title() : props.title());
                property.setDescription(props.description());
                property.setReadonly(props.readonly());
            }
        }
        Class<?> clazz = descriptor.getPropertyType();
        switch (property.getType()) {
            case STRING:
                // 如果字段使用DictValue注解，尝试使用字典值仓库
                if (null != field && field.isAnnotationPresent(DictValue.class)) {
                    DictValue dictValue = field.getAnnotation(DictValue.class);
                    property.prop("code", dictValue.value());
                }
            case LIST:
                // 尝试获取泛型参数，存在时，赋值子表单
                if (null != field && field.isAnnotationPresent(SubBean.class)) {
                    parseSubClass(field).ifPresent(subClazz -> property.setChildren(from(subClazz)));
                }
                break;
            case OBJECT:
                // 有子bean注解才处理
                if (null != field && field.isAnnotationPresent(SubBean.class)) {
                    property.setChildren(from(clazz));
                }
                break;
            case DB_REF:
                // 当存在db-ref时，解析为动态数据源
                Document document = AnnotationUtils.findAnnotation(clazz, Document.class);
                if (null != document) {
                    property.prop("uri", document.value());
                } else {
                    property.setType(BeanPropertyType.STRING);
                }
                break;
            case ENUM:
                // 为枚举时，填入枚举类型
                String name = StringFormats.camel2Line(ClassUtils.getShortClassName(clazz));
                property.prop("code", name);
                break;
        }
        return property;
    }

    /**
     * 类加载并解析属性
     *
     * @return 结果
     */
    public static List<BeanProperty> from(Class<?> clazz) {
        PropertyDescriptor[] descriptors = BeanUtils.getPropertyDescriptors(clazz);
        return Arrays.stream(descriptors)
                .filter(descriptor -> !"class".equals(descriptor.getName()))
                .map(descriptor -> BeanProperty.form(descriptor, clazz))
                .filter(property -> !property.isReadonly())
                .collect(Collectors.toList());
    }

    /**
     * 解析子类型，用于表格组件
     *
     * @param field 反射域
     * @return 结果
     */
    private static Optional<Class<?>> parseSubClass(Field field) {
        // 获取实际的泛型
        Type type = field.getGenericType();
        if (type instanceof ParameterizedType) {
            val types = TypeUtils.getTypeArguments((ParameterizedType) type);
            return types.entrySet().stream().findFirst().map(entry -> {
                Type parsed = entry.getValue();
                if (parsed instanceof Class) {
                    Class<?> clazz = (Class<?>) parsed;
                    if (!ClassUtils.isPrimitiveOrWrapper(clazz) && !CharSequence.class.isAssignableFrom(clazz)) {
                        return clazz;
                    }
                }
                return null;
            });
        }
        return Optional.empty();
    }

    public BeanProperty prop(String key, Object value) {
        if (null == props) {
            props = new HashMap<>();
        }
        props.put(key, value);
        return this;
    }
}
