package com.flyfish.framework.configuration.jwt;

import com.flyfish.framework.domain.base.IUser;
import com.flyfish.framework.utils.RedisOperations;
import io.jsonwebtoken.*;
import io.jsonwebtoken.io.Decoders;
import io.jsonwebtoken.io.DecodingException;
import io.jsonwebtoken.security.Keys;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.http.HttpHeaders;
import org.springframework.http.server.reactive.ServerHttpRequest;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.userdetails.User;
import org.springframework.web.server.ServerWebExchange;

import javax.annotation.Resource;
import java.security.Key;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * token提供者，负责生成token
 *
 * @author wangyu
 */
@Slf4j
public class TokenProvider implements InitializingBean {

    public static final String AUTHORIZATION_HEADER = "Authorization";
    private static final String AUTHORITIES_KEY = "auth";
    private final Boolean remember;
    private final long tokenValidityInMilliseconds;
    private final long tokenValidityInMillisecondsForRememberMe;

    @Resource
    private RedisOperations redisOperations;
    @Resource
    private JwtProperties jwtProperties;

    private Key key;

    public TokenProvider(JwtProperties jwtProperties) {
        this.remember = jwtProperties.isRemember();
        this.tokenValidityInMilliseconds = jwtProperties.getTokenValidityInSeconds() * 1000;
        this.tokenValidityInMillisecondsForRememberMe = jwtProperties.getTokenValidityInSecondsForRememberMe() * 1000;
    }

    @Override
    public void afterPropertiesSet() {
        byte[] keyBytes = Decoders.BASE64.decode(jwtProperties.getBase64Secret());
        this.key = Keys.hmacShaKeyFor(keyBytes);
    }

    /**
     * 从上下文获取token
     *
     * @return 结果
     */
    public Optional<String> retrieveToken(ServerWebExchange exchange) {
        ServerHttpRequest request = exchange.getRequest();
        String bearerToken = request.getHeaders().getFirst(AUTHORIZATION_HEADER);
        if (StringUtils.isNotBlank(bearerToken) && bearerToken.startsWith("Bearer ")) {
            return Optional.of(bearerToken.substring(7));
        }
        // 从cookies里头找
        if (MapUtils.isNotEmpty(request.getCookies())) {
            return request.getCookies().keySet().stream().filter(name -> name.equals(AUTHORIZATION_HEADER))
                    .findFirst().flatMap(name -> request.getCookies().get(name).stream()
                            .filter(value -> StringUtils.isNotBlank(value.getValue()) && value.getValue().startsWith("Bearer-"))
                            .findFirst()
                    ).map(cookie -> cookie.getValue().substring(7));
        }
        return Optional.empty();
    }

    public void addToken(ServerWebExchange exchange, Authentication authentication) {
        IUser user = (IUser) authentication.getPrincipal();
        String token = createToken(authentication, remember);
        HttpHeaders headers = exchange.getResponse().getHeaders();
        // app用户从头部返回，方便获取
        headers.add("Token", token);
        headers.add("Token-Valid-Time", String.valueOf(remember ? tokenValidityInMillisecondsForRememberMe :
                tokenValidityInMilliseconds));
        // token在web端的时间较短，不允许记住，所以使用短期
//        exchange.getResponse().addCookie(ResponseCookie.from(AUTHORIZATION_HEADER, "Bearer-" + token).
//                httpOnly(true).maxAge(tokenValidityInMilliseconds).build());
        // redis存储时间长
        redisOperations.set(token, true, tokenValidityInMillisecondsForRememberMe);
    }

    public void removeToken(ServerWebExchange exchange) {
        retrieveToken(exchange).ifPresent(token -> redisOperations.del(token));
    }

    public String createToken(Authentication authentication, boolean rememberMe) {
        String authorities = authentication.getAuthorities().stream()
                .map(GrantedAuthority::getAuthority)
                .collect(Collectors.joining(","));

        long now = (new Date()).getTime();
        Date validity;
        if (rememberMe) {
            validity = new Date(now + this.tokenValidityInMillisecondsForRememberMe);
        } else {
            validity = new Date(now + this.tokenValidityInMilliseconds);
        }

        return Jwts.builder()
                .setSubject(authentication.getName())
                .claim(AUTHORITIES_KEY, authorities)
                .signWith(key, SignatureAlgorithm.HS512)
                .setExpiration(validity)
                .compact();
    }

    public Authentication getAuthentication(String token) {
        Claims claims = Jwts.parserBuilder()
                .setSigningKey(key)
                .build()
                .parseClaimsJws(token)
                .getBody();

        Collection<? extends GrantedAuthority> authorities =
                Arrays.stream(claims.get(AUTHORITIES_KEY).toString().split(","))
                        .filter(StringUtils::isNotBlank)
                        .map(SimpleGrantedAuthority::new)
                        .collect(Collectors.toList());

        User principal = new User(claims.getSubject(), "", authorities);
        return new UsernamePasswordAuthenticationToken(principal, token, authorities);
    }

    public boolean validateToken(String authToken) {
        try {
            Jwts.parserBuilder().setSigningKey(key).build().parseClaimsJws(authToken);
            return true;
        } catch (io.jsonwebtoken.security.SecurityException | MalformedJwtException e) {
            log.info("Invalid JWT signature.");
            log.trace("Invalid JWT signature trace: {}", e, e);
        } catch (ExpiredJwtException e) {
            log.info("Expired JWT token.");
            log.trace("Expired JWT token trace: {}", e, e);
        } catch (UnsupportedJwtException e) {
            log.info("Unsupported JWT token.");
            log.trace("Unsupported JWT token trace: {}", e, e);
        } catch (IllegalArgumentException e) {
            log.info("JWT token compact of handler are invalid.");
            log.trace("JWT token compact of handler are invalid trace: {}", e, e);
        } catch (DecodingException e) {
            log.info("Token解析失败！");
            log.trace("Token解析失败: {}", e, e);
        }
        return false;
    }
}
