package com.flyfish.framework.service;

import com.flyfish.framework.domain.base.Qo;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.List;

/**
 * 基础Service，全异步接口
 * 代理Repo的所有方法，内部封装常用的全部操作
 *
 * @author wybab
 */
public interface BaseReactiveService<T> {

    /**
     * 查询
     *
     * @param entity 实体本身
     * @return 结果
     */
    Mono<T> getOne(T entity);

    /**
     * 查询
     *
     * @param query 查询模型
     * @return 结果
     */
    Mono<T> getOne(Qo<T> query);

    /**
     * 通过Id查询
     *
     * @param id 主键
     * @return 结果
     */
    Mono<T> getById(String id);

    /**
     * 根据ID集合来查询
     *
     * @param ids id列表
     * @return 结果
     */
    Flux<T> getByIds(List<String> ids);

    /**
     * 查询列表
     *
     * @param entity 实体本身
     * @return 结果
     */
    Flux<T> getList(T entity);

    /**
     * 查询列表
     *
     * @param query 查询
     * @return 结果
     */
    Flux<T> getList(Qo<T> query);

    /**
     * 分页查询列表
     *
     * @param entity 实体本身
     * @return 结果
     */
    Mono<Page<T>> getPageList(T entity, Pageable pageable);

    /**
     * 分页查询列表
     *
     * @param query 查询实体
     * @return 结果
     */
    Mono<Page<T>> getPageList(Qo<T> query);

    /**
     * 获取所有对象
     *
     * @return 结果
     */
    Flux<T> getAll();

    /**
     * 查询总记录数
     *
     * @return 总数
     */
    Mono<Long> countAll();

    /**
     * 查询总记录数
     *
     * @param entity 通过实体查询个数
     * @return 结果
     */
    Mono<Long> count(T entity);

    /**
     * 插入新记录
     *
     * @param entity 实体
     */
    Mono<T> create(T entity);

    /**
     * 插入 不插入null字段
     *
     * @param entity 实体
     */
    Mono<T> createSelective(T entity);

    /**
     * 删除
     *
     * @param entity 实体
     */
    Mono<Void> delete(T entity);

    /**
     * 根据Id删除
     *
     * @param id 主键
     */
    Mono<Void> deleteById(String id);

    /**
     * 根据id更新
     *
     * @param entity 实体
     */
    Mono<T> updateById(T entity);

    /**
     * 不update null
     *
     * @param entity 实体
     */
    Mono<T> updateSelectiveById(T entity);

    /**
     * 根据ID集合批量删除
     *
     * @param ids id结合
     */
    Mono<Void> deleteBatchByIds(List<String> ids);

    /**
     * 批量更新
     *
     * @param entities 要批量更新的集合
     */
    Flux<T> updateBatch(List<T> entities);
}
